#ifndef SURFACE_FACTORY
#define SURFACE_FACTORY 1

#include "Graphs.h"
#include "Notifications.h"

#include <map>
#include <set>
#include <string>

class ImageProvider;
class ImageUser;
class ImageFactory;

class ImageUser:public Notifier {
protected:
    friend class ImageProvider;
    ImageProvider * producer;
    
    Surface * current;
    int serial;
    
public:
    ImageUser(TaskQueue * tq = InternalTask::syncTaskQueue);
    virtual ~ImageUser();
    
    void bind(ImageProvider * nvproducer);
    
    bool binded() const;/*X*/
    ImageProvider * getProducer() const;/*X*/
    // Check if a new surface is available.
    Surface * getSurface();
    Surface * getFinalSurface();
    int getSerial() const;
};

/**
 * Fourni une ressource.
 * Trace les utilisateurs de cette ressource
 */
class ImageProvider : public InternalTask{
    friend class ImageUser;
    friend class ImageFactory;
    
protected:
    std::string name;
    
    /// Actually update the surface.
    virtual void run(TaskNotification * tn);
    
    /// Usage count for each Surface created.
    /// Modified only under global lock
    std::map<Surface*,int> usageCount;
    
    /// List of current users
    /// Modified only under global lock
    std::set<ImageUser*> users;
    
    void addUser(ImageUser * i);
    void removeUser(ImageUser * i);
    
    /// Current available image
    /// Modified only under global lock
    Surface * currentImage;
    
    // Change currentImage within lock
    void changeSurface(Surface * s);
    
    // Called without global lock
    Surface * updateSurface(Surface * actual, bool & changed);
    
    // Called without global lock
    void releaseSurface(Surface * actual);
    
    // Command a reload of the image...
    virtual void reload();
    
    /// Call only under globallock : return true if ImageProvider is unused
    bool used();
public:
    ImageProvider(std::string name);
    virtual ~ImageProvider();
    
    /// Call without lock
    bool isDone();
    
    std::string getName() const;
    
    bool done;
};

class ImageIdentifier {
    
public:
    bool isThumb;
    int sx,sy;
    std::string name;
    
    ImageIdentifier(const std::string & vname = "noname") {
	name = vname;
	isThumb = false;
    }
	
    ImageIdentifier(const std::string & vname,int vsx,int vsy) {
	name = vname;
	isThumb = true;
	sx = vsx;
	sy = vsy;
    }
    
    bool operator<(const ImageIdentifier t) const{
	if (t.name != name) {
	    return t.name < name;
	}
	if (isThumb != t.isThumb) {
	    return isThumb < t.isThumb;
	}
	if (!isThumb) {
	    return false;
	}
	if (sx != t.sx) {
	    return sx < t.sx;
	}
	if (sy != t.sy) {
	    return sy < t.sy;
	}
	return false;
    }
};

class ImageFactory {
    friend class ImageProvider;
public:
    static ImageFactory * factory;
private:
    
    /// Only modified under globallock
    std::map<ImageIdentifier,ImageProvider*> requests;
    
    //ImageProvider * findNextTask();
    void doLoad(ImageIdentifier path,ImageUser & user);/*X*/
    
public:
    ImageFactory();
    
    void cleanUp(); // Called within globallock
    static void invalidateImage(const std::string & name);
    static void loadImage(std::string,ImageUser & user);
    static void loadThumb(std::string,int sx,int sy,ImageUser & user);
};


#endif
