#include <SDL.h>
#include <SDL_thread.h>
#include <SDL_image.h>

#include <stdexcept>

#include "ImageFactory.h"

ImageProvider::ImageProvider(std::string vname) : InternalTask(InternalTask::asyncTaskQueue), name(vname), usageCount()
{
    jsrlog("%#x: created image provider for %s\n",this,vname.c_str());
    currentImage = new Surface();
    done = false;
}

ImageProvider::~ImageProvider()
{
    if (!usageCount.empty()) {
	throw std::runtime_error("Destroying imageProvider still in use");
    }
    delete(currentImage);
}

std::string ImageProvider::getName() const
{
    return name;
}

bool ImageProvider::isDone()
{
    bool rslt;
    InternalTask::lock();
    rslt = done;
    InternalTask::unlock();
    return rslt;
}

void ImageProvider::run(TaskNotification * tn)
{
    if (tn) delete(tn);
    InternalTask::unlock();
    jsrlog("%#x : loading %s\n",this,name.c_str());
    Surface * nv = new Surface();
    nv->LoadImage(JSReflection::decodeBinaryPath(name).c_str(),Surface::Alpha);
    InternalTask::lock();
    changeSurface(nv);
    done = true;
}

void ImageProvider::addUser(ImageUser * i)
{
    InternalTask::lock();
    i->waitNotification(this);
    users.insert(i);
    InternalTask::unlock();
}

void ImageProvider::removeUser(ImageUser * i)
{
    InternalTask::lock();
    i->unwaitNotification(this);
    users.erase(i);
    InternalTask::unlock();
}

void ImageProvider::reload()
{
    InternalTask::lock();
    done = false;
    autoSignal();
    InternalTask::unlock();
}

/**
 * Called by ImageUser, to update their image. 
 * 
 * If ImageUser is outofdate, its surface is released/destroyed
 */
Surface * ImageProvider::updateSurface(Surface * actual, bool & changed)
{
    InternalTask::lock();
    
    if (actual == currentImage) {
	changed = false;
	InternalTask::unlock();
	return actual;
    }
    changed = true;
    if (actual) {
	if (usageCount.find(actual) != usageCount.end()) {
	    //jsrlog("dec usageCount[%#x] to %d\n",actual,usageCount[actual] - 1);
	    usageCount[actual]--;
	    if (!usageCount[actual]) {
		delete(actual);
		usageCount.erase(actual);
	    }
	}
    }
    actual = currentImage;
    if (actual) {
	//jsrlog("inc usageCount[%#x] to %d\n",actual,usageCount[actual] + 1);
	usageCount[actual]++;
    }
    
    InternalTask::unlock();
    return actual;
}

bool ImageProvider::used()
{
    for(typeof(usageCount.begin()) it = usageCount.begin(); it != usageCount.end(); ++it) {
	//jsrlog("usage for %#x is %d\n",it->first,it->second);
	if ((it->second) > 0) {
	    return true;
	}
    }
    return false;
}

void ImageProvider::releaseSurface(Surface * actual)
{
    InternalTask::lock();
    if (actual) {
	if (usageCount.find(actual) != usageCount.end()) {
	    
	    usageCount[actual]--;

	    if ((actual != currentImage) && (!usageCount[actual])) {
		delete(actual);
	    }
	    if (!usageCount[actual]) {
		usageCount.erase(actual);
	    }
	}
    }
    InternalTask::unlock();
}

void ImageProvider::changeSurface(Surface * s)
{
    InternalTask::lock();
    if (currentImage) {
	if (usageCount[currentImage] == 0) {
	    delete(currentImage);
	    usageCount.erase(currentImage);
	}
    }
    currentImage = s;
    emitNotification();
    InternalTask::unlock();
}


