#ifndef TASK_H
#define TASK_H 1

#include <set>
#include <list>
#include <map>
#include <SDL_thread.h>

#define TASK_STOPPED	0			// No notification pending
#define TASK_RUNNING	1			// Currently running
#define TASK_READY	2			// Has pending notification
#define TASK_WAITING	3			// Is waiting for notification

class InternalTask;

class TaskQueue {
    std::list<InternalTask*> ** tasks;
    
    bool signaled;
    
    int priority;
    
    virtual void wakeup() = 0;
public:
    TaskQueue();
    virtual ~TaskQueue();
    
    
    void setState(InternalTask * t,int state); 	// Must be called within lock
    
    void taskRun(InternalTask * t);		// Must be called within lock
    
    InternalTask * nextReady();	 	  	// Return next ready task. Must own the lock
    
    bool step();				// Must be called without lock. Return true if something was done
};

class TaskNotification {
public:
    InternalTask * from, * to;
    TaskNotification(InternalTask * from, InternalTask * to);
    ~TaskNotification();
};

class InternalTask {
    friend class TaskQueue;
    int state;					// Task may be 0:nothing, 1:Running, 2:Ready, 3:Waiting
    int priority;
    std::set < InternalTask* > waiting;		// Set of task waiting notification from this one
    std::set < InternalTask* > waitedFor;	// Set of task which will send notification to this
    
    std::map < InternalTask *, TaskNotification* > emitedNotification;
                                                // List of notification
    std::map < InternalTask *, TaskNotification* > pendingNotification;
                                                // Pending Notification. Must be dequeued within run
    
    TaskQueue * taskQueue;
protected:
    void waitNotification(InternalTask * t);	// Must be called within lock
    void unwaitNotification(InternalTask * t);	// Remove t from waiting and discard any pending notification.
                                                // Must be called within lock
    void emitNotification();			// Wake up any sleeping tasks. Must be called within the lock
    void createNotification(InternalTask * to);// Create a new notification. Must be called within the lock
public:
    InternalTask(TaskQueue * tq);
    virtual ~InternalTask();
    
    void setPriority(int p);
    
    void start();				// Just put in the ready state. Must be locked
    void autoSignal();				// Task signal itsefl. Must be locked
    
    virtual bool suicide();			/// Called after task run. If return true, object is deleted
    virtual void run(TaskNotification * tn);	// When a task starts to run, it's waiting queue is empty 
    bool isRunning();				/// Only available under globallock
    bool pendingSignal();
    
    static void lock();
    static void unlock();
    static void condWait(SDL_cond * cond);
    static TaskQueue * newAsyncTaskQueue(int threadnb);
    static void startAsyncThreads(int nb);
    static void initSyncThread(int fd);
    
    static bool exitFlag();			// Must be called locked
    
    static TaskQueue * asyncTaskQueue;
    static TaskQueue * syncTaskQueue;
};


#endif
