#include <SDL.h>
#include <SDL_thread.h>

#include <stdexcept>

#include "Graphs.h"

#include "ThumbProvider.h"

std::list<ThumbProvider*> running; 
std::list<ThumbProvider*> queue;

void ThumbProvider::nextQueued() {
    ThumbProvider * tp;
    if (running.size() > 2) {
	return;
    }
    if (queue.empty()) {
	return;
    }
    tp = queue.front();
    queue.pop_front();
    running.push_back(tp);
    tp->autoSignal();
}

ThumbProvider::ThumbProvider(std::string filename,int sx,int sy) : ImageProvider(filename),ImageUser(InternalTask::asyncTaskQueue),meta(InternalTask::asyncTaskQueue)
{
    sizex=sx;
    sizey=sy;
    this->setPriority(-1);
    meta.setTarget(this);
}

ThumbProvider::~ThumbProvider()
{
    InternalTask::lock();
    queue.remove(this);
    running.remove(this);
    nextQueued();
    InternalTask::unlock();
}

void ThumbProvider::init() 
{
    InternalTask::lock();
    queue.push_back(this);
    nextQueued();
    InternalTask::unlock();
}

void ThumbProvider::notified(Notifier*)
{
    this->autoSignal();
}

void ThumbProvider::reload()
{
    InternalTask::lock();
    done = false;
    this->autoSignal();
    InternalTask::unlock();
}

void ThumbProvider::run(TaskNotification * tn)
{
    std::string binaryImage;
    std::string thumbName;
    char buffer[256];
    if (tn) delete(tn);
    if (!meta.binded()) {
	meta.bind(name);
	if (!meta.isReady()) {
	    return;
	}
    }
    snprintf(buffer,256,"%dx%d",sizex,sizey);
    thumbName = std::string("thumb") + buffer;
    if (meta.getKey(thumbName,binaryImage)) {
	InternalTask::unlock();
	Surface * nv = new Surface();
	nv->decode(binaryImage);
	InternalTask::lock();
	this->emitNotification();
	changeSurface(nv);
	done = true;
	meta.unbind();
	jsrlog("Thumb %s provided from metadata.\n",name.c_str());
	bind(0);
	running.remove(this);
	nextQueued();
	return;
    }
    jsrlog("Key %s not found in metadata.\n",thumbName.c_str());
    // Need image...
    if (!binded()) {
	jsrlog("%#x : bind image : %s\n",this,name.c_str());
	ImageFactory::loadImage(name,*this);
	ImageProvider * ip = getProducer();
	jsrlog("%#x : bind to : %#x\n",(ImageProvider *)this,(ImageProvider *)ip);
	if (!ip->done) {
	    jsrlog("Waiting for image\n");
	    return;
	}
    }
    
    InternalTask::unlock();
    jsrlog("ThumbProvider(%#x)::load\n",this);
    Surface * src;
    if (src = getFinalSurface()) {
	jsrlog("Image is ready\n");
	// Use src
	Surface * nv = new Surface();
	
	if ((src->GetWidth() > 0) && (src->GetHeight() > 0)) {
	    int sx,sy;
	    int pt[4][2] = {
		    {0,0},
		    {0,256*src->GetHeight()},
		    {256*src->GetWidth(),0},
		    {256*src->GetWidth(),256*src->GetHeight()}};
	
	    sx = src->GetWidth() * sizex / src->GetWidth();
	    sy = src->GetHeight() * sizex / src->GetWidth();
	    if (sy > sizey) {
		sx = src->GetWidth() * sizey / src->GetHeight();
		sy = src->GetHeight() * sizey / src->GetHeight();
	    }
	    
	    // No alpha for thumb!
	    nv->Create(sx,sy,
		    Surface::Opaque,
		    //(src->HasTransparency()?Surface::Alpha:Surface::Opaque),
		    false,0);
		    
	    SDL_stretch(src->GetSDLSurface(),nv->GetSDLSurface(),pt);
	    //nv->LoadImage(JSReflection::decodeBinaryPath(path).c_str(),Surface::Alpha);
	}
	InternalTask::lock();
	
	
	this->emitNotification();
	changeSurface(nv);
	done = true;
	bind(0);
	running.remove(this);
	
	nextQueued();
	
	
	// save back the thumb
	if (meta.binded()) {
	    // FIXME : nv can be destroyed from here ! so we need to do it locked !
	    nv->encode(binaryImage);

	    meta.setKey(thumbName,binaryImage);
	}
	
    } else {
	InternalTask::lock();
    }
}


