#include <SDL.h>
#include <SDL_image.h>

#include "Graphs.h"
#include "Application.h"
#include "Widget.h"
#include "Layout.h"
#include "Style.h"
#include <string>
#include <math.h>

/*
Widget relation(s) :
	0-1 	father,childs
	* 	InputEventHandler
	1	Layout
	1(?)	Font
	1	ImageUser
*/

Widget::Widget() : JSReflection::Object(Application::getJsContext()), text()
{
    father = 0;
    layout = 0;
    childShiftX = 0;
    childShiftY = 0;
    visible = false;
    PreserveCounterpart = true;
    backgroundAsThumb = false;
    x = 0;
    y = 0;
    w = 1;
    h = 1;
    textFont = 0;
    layedOut = 0;
    
    textHAlign = -1;
    textVAlign = -1;

    backgroundVAlign = -1;
    backgroundHAlign = -1;
    
    roundShape = 0;
    modalFor = 0;
    background.setTarget(this);
}

Widget::~Widget()
{
    std::list<Widget*>::const_iterator cit;
    
    // FIXME : fonts are not release.
    while((cit = childs.begin()) != childs.end()) {
	delete(*cit);
    }
    if (modalFor) {
	modalFor->modals.remove(this);
    }
    
    for(typeof(modals.begin()) it = modals.begin(); it != modals.end(); ++it) {
	(*it)->modalFor = 0;
    }
    
    if (father) {
	father->childs.remove(this);
	father->looseFatherData(this);
    }
    if (layout) {
	delete (layout);
    }
}

void Widget::setLayout(Layout * l)
{
    if (layout) {
	delete layout;
    }
    
    layout = l;
    assert(!layout->widget);
    layout->widget = this;
}

void Widget::updateLayout()
{
    if (!layout) { return; }
    layout->update();
    for(std::list<Widget*>::const_iterator cit = childs.begin();cit != childs.end(); ++cit) {
	layout->relayoutWidget(*cit);
	(*cit)->updateLayout();
    }
}

void Widget::destroy()
{
    if (checkScreenUpdate()&& isGlobalVisible()) {
	askScreenUpdate();
    }

    removeChilds();
    delete(this);
}

void Widget::setFather(Widget * vfather)
{
    if (vfather == this) vfather = 0;
    if (father == vfather) return;

    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }

    if (father) {
    	father->childs.remove(this);
	father->looseFatherData(this);
    }
    layedOut = false;
    
    father = vfather;
    
    if (father) {
    	father->childs.push_back(this);
	father->initFatherData(this);
	if (father->layout) {
	    father->layout->addWidget(this);
	}
    }

    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

Widget * Widget::getFather() const
{
    return father;
}

void Widget::setBackgroundAsThumb(bool bat)
{
    if (backgroundAsThumb != bat) {
	backgroundAsThumb = bat;
    
	backgroundBind();
    }
}

bool Widget::getBackgroundAsThumb() const 
{
    return backgroundAsThumb;
}

void Widget::setPosition(int vx,int vy,int vw,int vh)
{
    bool sizechanged;
    if ((x == vx) && (y == vy) && (w == vw) && (h == vh)) {
	return;
    }
    x = vx;
    y = vy;
    sizechanged = (w!=vw)||(h!=vh);
    w = vw;
    h = vh;
    
    if (sizechanged && backgroundAsThumb) {
	backgroundBind();
    }
    
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

void Widget::setRoundShape(int i)
{
    if (roundShape == i) return;
    roundShape = i;
    
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

int Widget::getRoundShape() const
{
    return roundShape;
}

void Widget::setColor(unsigned int c)
{
    color.color = (color.color & 0xff) | ((Uint32)c) << 8;
    
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

unsigned int Widget::getColor() const
{
    return color.color >> 8;
}

void Widget::setAlpha(unsigned int c)
{
    if (c>255) c = 255;
    color.color = (color.color & 0xffffff00) | ((Uint32) c);

    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

unsigned int Widget::getAlpha() const
{
    return color.color & 0xff;
}

void Widget::setColor(const Color& c)
{
    color = c;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

void Widget::setColor(int r,int g,int b)
{
    color.setR(r);
    color.setG(g);
    color.setB(b);
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

void Widget::setTextColor(unsigned int c)
{
    textColor.color = (textColor.color & 0xff) | ((Uint32)c) << 8;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

unsigned int Widget::getTextColor() const
{
    return textColor.color >> 8;
}

bool Widget::getVisible() const
{
    return visible;
}

void Widget::setVisible(bool b)
{
    bool wasGlobalVisible;
    if (visible == b) {
	return;
    }
    wasGlobalVisible = isGlobalVisible();
    
    if (checkScreenUpdate() && wasGlobalVisible) {
	askScreenUpdate();
    }

    visible = b;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
    
    if (isGlobalVisible() != wasGlobalVisible) {
	// Report visibility change to all childs
	notifyGlobalVisibleChanged();
    }
}

void Widget::backgroundBind()
{
    if (isGlobalVisible()) {
	if (background_file == "") {
	    background.bind(0);
	} else {
	    if (backgroundAsThumb) {
		ImageFactory::loadThumb(background_file,w,h,background);
	    } else {
		ImageFactory::loadImage(background_file,background);
	    }
	}
	
	// FIXME : update when image is loaded as well !!!
	if (checkScreenUpdate() && isGlobalVisible()) {
	    askScreenUpdate();
	}
    } else {
	background.bind(0);
    }
}

void Widget::globalVisibleChanged()
{
    backgroundBind();
}

void Widget::notifyGlobalVisibleChanged()
{
    globalVisibleChanged();
    // Report the notification to all visible childs
    for(typeof(childs.begin()) cit = childs.begin(); cit != childs.end(); ++cit) {
	if ((*cit)->visible) {
	    (*cit)->notifyGlobalVisibleChanged();
	}
    }
}

int Widget::getTop() const {
    return y;
}

void Widget::setTop(int u) {
    if (y == u) return;
    y = u;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }

}

int Widget::getLeft() const {
    return x;
}

void Widget::setLeft(int u) {
    if (x == u) return;

    x = u;
    
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }

}

int Widget::getWidth() const {
    return w;
}

void Widget::setWidth(int u) {
    if (w == u) return;
    w = u;
    if (backgroundAsThumb) {
	backgroundBind();
    }
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

int Widget::getHeight() const {
    return h;
}

void Widget::setHeight(int u) {
    if (h == u) return;
    h = u;
    if (backgroundAsThumb) {
	backgroundBind();
    }
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

void Widget::setText(const std::string & s) {
    if (text == s) return;
    text = s;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }

}

const std::string & Widget::getText() const
{
    return text;
}

void Widget::setStyle(const std::string & s) {
    style = s;
    Style::applyStyle(this);
}

const std::string & Widget::getStyle() const {
    return style;
}

const std::string & Widget::getSubStyle() const {
    return substyle;
}

void Widget::subStyleAdd(const std::string & name) {
    unsigned pos;
    std::string search(":");
    search+=name;
    search+=":";
    
    if ((pos = substyle.find(search)) != std::string::npos) {
	return;
    }
    if (substyle.length() == 0) {
	substyle+=":";
    }
    substyle+=name+":";
    Style::applyStyle(this);
}

void Widget::subStyleRemove(const std::string & name) {
    unsigned pos;
    std::string search(":");
    search+=name;
    search+=":";
    
    if ((pos = substyle.find(search)) == std::string::npos) {
	return;
    }
    
    substyle.erase(pos,search.length() - 1);
    
    Style::applyStyle(this);
}

bool Widget::hasSubStyle(const std::string & name) const{
    return substyle.find(std::string(":")+name+":") != std::string::npos;
}

void Widget::setSubStyle(const std::string & name,bool value) {
    bool hasit = hasSubStyle(name);
    if (hasit == value) { return; }
    if (value) {
	subStyleAdd(name);
    } else {
	subStyleRemove(name);
    }
}

void Widget::setGlobalName(const std::string & s) {
     GetContext()->GetGlobal()->setProperty(s.c_str(),this);
}

std::string Widget::getGlobalName() const
{
    return "";
}

void Widget::setBackground(const std::string & s) {
    background_file = s;
    backgroundBind();
}

std::string Widget::getBackground() const {
    
    /*if (background.binded()) {
	return background.getProducer()->getName();
    } */
    return background_file;
}

int Widget::getBackgroundVAlign() const{
    return backgroundVAlign;
}

void Widget::setBackgroundVAlign(int i) {
    if (backgroundVAlign == i) return;
    backgroundVAlign = i;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

int Widget::getBackgroundHAlign() const{
    return backgroundHAlign;
}

void Widget::setBackgroundHAlign(int i) {
    if (backgroundHAlign == i) return;
    backgroundHAlign  = i;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}


void Widget::setTextFont(Font * f) {
    if (textFont == f) return;
    textFont = f;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

int Widget::getTextVAlign() const{
    return textVAlign;
}

void Widget::setTextVAlign(int i) {
    if (textVAlign == i) return;
    textVAlign = i;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

int Widget::getTextHAlign() const{
    return textHAlign;
}

void Widget::setTextHAlign(int i) {
    if (textHAlign == i) return;
    textHAlign  = i;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

void Widget::notified(Notifier * n)
{
    jsrlog("Widget %#x notified by %#x\n",this,n);
    if (n != &background) return;
    
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

Font * Widget::getTextFont() const {
    return textFont;
}


std::string Widget::debug(int n) const
{
    std::string result;
    char buffer[2048];
    for(int i=0;i<n;++i) {
	result+=" ";
    }
    snprintf(buffer,2048,"<%s %c %dx%d@(%d,%d) %#x",GetClassName(),(visible?'.':'!'),w,h,x,y,color.color);
    result+=buffer;
    if (childs.empty()) {
	result+=">";
    } else {
	
	for(std::list<Widget*>::const_iterator cit = childs.begin();cit != childs.end(); ++cit) {
	    result+="\n";
	    result+=(*cit)->debug(n+1);
	}
	result+=">";
    }
    return result;
}

void Widget::draw(Surface * dest,const ScreenArea * clip)
{
    if (!visible) {
	jsrlog("Widget::draw called on invisible widget\n");
	return;
    }
    Painter p(this,dest,clip);
    
    if (color.a()) {
	int vroundShape = roundShape;

	if (2 * vroundShape > getWidth()) {
	    vroundShape = getWidth() / 2;
	}
	    
	if (2 * vroundShape > getHeight()) {
	    vroundShape = getHeight() / 2;
	}
	    
	if (vroundShape) {
	    int cy;
	    for(cy = 0; cy < vroundShape - 1; ++cy) {
		double pos = 1.0 - ((cy+1) *1.0 / vroundShape);
		int cx = (int)(vroundShape * (1.0 - sqrt(1.0 - pos * pos)));
		p.FillRectangle(cx,cy ,w - cx * 2,1,color);
		p.FillRectangle(cx,h - 1 - cy ,w - cx * 2,1,color);
	    }
	    
	    p.FillRectangle(0,cy ,w,h  - 2 * (vroundShape - 1),color);
	} else {
	    p.FillRectangle(0,0,w,h,color);
	}
    }
    
    
    Surface * backsurf = background.getSurface();
    if (backsurf) {
	int drawx,drawy;
	if (backgroundHAlign < 0) {
	    drawx = 0;
	} else if (backgroundHAlign > 0) {
	    drawx = getWidth() - backsurf->GetWidth();
	} else {
	    drawx = (getWidth() - backsurf->GetWidth()) / 2;
	}
	
	if (backgroundVAlign < 0) {
	    drawy = 0;
	} else if (backgroundVAlign > 0) {
	    drawy = getHeight() - backsurf->GetHeight();
	} else {
	    drawy = (getHeight() - backsurf->GetHeight()) / 2;
	}
	
	p.Blit(backsurf,drawx,drawy);
    }
    
    if (textFont && text.size()) {
	int drawx,drawy;
	if (textHAlign < 0) {
	    drawx = 0;
	} else if (textHAlign > 0) {
	    drawx = getWidth() - 1;
	} else {
	    drawx = getWidth() / 2;
	}
	if (textVAlign < 0) {
	    drawy = 0;
	} else if (textVAlign > 0) {
	    drawy = getHeight() - 1;
	} else {
	    drawy = getHeight() / 2;
	}
	
	p.DrawText(text.c_str(),drawx,drawy,textFont,textColor,textHAlign,textVAlign);
    }
    
    if (childs.empty()) return;
    
    for(std::list<Widget*>::const_iterator cit = childs.begin();cit != childs.end(); ++cit) {
	if (!(*cit)->visible) continue;
	(*cit)->draw(dest,p.getArea());
    }
}

void Widget::grabInput(Widget * other)
{
    if (modalFor == other) return;
    
    if (modalFor) {
	modalFor->modals.remove(this);
    }
    
    modalFor = other;
    if (other) {
	modalFor->modals.push_front(this);
    }
}

void Widget::ungrabInput()
{
    grabInput(0);
}

void Widget::getAbsolutePos(int & absx,int & absy)
{
    Widget * cur = this->father;
    absx = x;
    absy = y;
    while(cur) {
        absx += cur->x - cur->childShiftX;
        absy += cur->y - cur->childShiftY;

        cur = cur->father;
    }
}

void Widget::addInputEventHandler(InputEventHandler * ieh)
{
    inputs.push_back(ieh);
}

void Widget::removeInputEventHandler(InputEventHandler * ieh)
{
    inputs.remove(ieh);
}

// visibleOnly == only ask to visible childs...
bool Widget::handleInputEvent(const std::string & name,bool visibleOnly)
{
    if (visibleOnly && !isGlobalVisible()) {
	return false;
    }

    if (!modals.empty()) {
	Widget * modal = (*modals.begin());
	// Give it to the modal
	if ((!visibleOnly) || modal->isGlobalVisible()) {
	    if (modal->handleInputEvent(name,modal->isGlobalVisible())) {
		return true;
	    }
	}
	return false;
    }
    
    // Give it to the childs...
    for(std::list<Widget*>::const_iterator cit = childs.begin(); cit != childs.end(); ++cit) {
	if ((!visibleOnly) || (*cit)->isGlobalVisible()) {
	    if ((*cit)->handleInputEvent(name,(*cit)->isGlobalVisible())) {
		return true;
	    }
	}
    }
    // Then try handlers...
    for(std::list<InputEventHandler*>::const_iterator cit = inputs.begin(); cit != inputs.end(); ++cit) {
	if ((*cit)->match(name,isGlobalVisible())) {
	    (*cit)->call(name);
	    return true;
	}
    }
    return false;
}

void Widget::removeChilds()
{
    while(!childs.empty()) {
	delete(*(childs.begin()));
    }
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

void Widget::initFatherData(Widget * child)
{
}

void Widget::looseFatherData(Widget * child)
{
}

bool Widget::checkScreenUpdate() const
{
    return (!Application::current->updater.programmed());
}

bool Widget::isGlobalVisible() const
{
    const Widget * w = this;
    
    while(w->visible) {
	w = w->father;
	if (!w) {
	    return true;
	}
    }
    
    return false;
}

void Widget::askScreenUpdate() const
{
    Application::current->updater.ask(true);
}

Widget * newWidget()
{
    return new Widget();
}

int debugWidget()
{
    printf("%s\n",Application::current->getRoot()->debug(0).c_str());
    return 0;
}

JS_EXPORT_CLASS(Widget,JSReflection::Object)
JS_RW_PROPERTY(Widget,father,getFather,setFather,Widget*,"%x",Widget::ClassInfo);
JS_RW_PROPERTY(Widget,top,getTop,setTop,int,"%i",0);
JS_RW_PROPERTY(Widget,left,getLeft,setLeft,int,"%i",0);
JS_RW_PROPERTY(Widget,width,getWidth,setWidth,int,"%i",0);
JS_RW_PROPERTY(Widget,height,getHeight,setHeight,int,"%i",0);
JS_RW_PROPERTY(Widget,visible,getVisible,setVisible,bool,"%b",0);
JS_RW_PROPERTY(Widget,roundShape,getRoundShape,setRoundShape,int,"%i",0);
JS_RW_PROPERTY(Widget,color,getColor,setColor,int,"%i",0);
JS_RW_PROPERTY(Widget,alpha,getAlpha,setAlpha,int,"%i",0);

JS_RW_PROPERTY(Widget,background,getBackground,setBackground,JSReflection::JS_String,"%s",0);
JS_RW_PROPERTY(Widget,backgroundHAlign,getBackgroundHAlign,setBackgroundHAlign,int,"%i",0);
JS_RW_PROPERTY(Widget,backgroundAsThumb,getBackgroundAsThumb,setBackgroundAsThumb,bool,"%b",0);
JS_RW_PROPERTY(Widget,backgroundVAlign,getBackgroundVAlign,setBackgroundVAlign,int,"%i",0);

JS_RW_PROPERTY(Widget,textColor,getTextColor,setTextColor,int,"%i",0);
JS_RW_PROPERTY(Widget,text,getText,setText,JSReflection::JS_String,"%s",0);
JS_RW_PROPERTY(Widget,textFont,getTextFont,setTextFont,Font*,"%x",Font::ClassInfo);
JS_RW_PROPERTY(Widget,textHAlign,getTextHAlign,setTextHAlign,int,"%i",0);
JS_RW_PROPERTY(Widget,textVAlign,getTextVAlign,setTextVAlign,int,"%i",0);

JS_RW_PROPERTY(Widget,style,getStyle,setStyle,JSReflection::JS_String,"%s",0);
JS_RO_PROPERTY(Widget,substyle,getSubStyle,JSReflection::JS_String,"%s",0);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_1(Widget,subStyleAdd,JSReflection::JS_String,"%s",0);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_1(Widget,subStyleRemove,JSReflection::JS_String,"%s",0);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_2(Widget,setSubStyle,JSReflection::JS_String,"%s",0,bool,"%b",0);

JS_RW_PROPERTY(Widget,globalname,getGlobalName,setGlobalName,JSReflection::JS_String,"%s",0);

JS_NOCONTEXT_OBJECT_FUNCTION_1(Widget,loadXml,bool,"%b",0,JSReflection::JS_String,"%s",0);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_4(Widget,setPosition,int,"%i",0,int,"%i",0,int,"%i",0,int,"%i",0);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_3(Widget,setColor,int,"%i",0,int,"%i",0,int,"%i",0);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_1(Widget,setFather,Widget*,"%x",Widget::ClassInfo);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_1(Widget,grabInput,Widget*,"%x",Widget::ClassInfo);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_0(Widget,ungrabInput);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_1(Widget,setVisible,bool,"%b",0);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_0(Widget,removeChilds);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_0(Widget,destroy);
JS_NOCONTEXT_OBJECT_VOID_FUNCTION_0(Widget,updateLayout);
JS_NOCONTEXT_GLOBAL_FUNCTION_0(debugWidget,int,"%i",0);
JS_NOCONTEXT_GLOBAL_FUNCTION_0(newWidget,Widget*,"%x",Widget::ClassInfo);

