#include <libxml/parser.h>
#include <libxml/parserInternals.h>

#include <stdexcept>

#include "jsreflect.h"
#include "XmlParser.h"


static std::string xmlChar2string(const xmlChar* value, int len = -1)
{
    int l = (len != -1 ? len : strlen((const char*)value));
    std::string str((const char*)value,l);
    return str;
/*    return std::string((const char*)value);
   
    
    int outlen = l + 256;
    unsigned char buffer[outlen + 1];
    int rslt;
    
    rslt = UTF8Toisolat1(buffer,&outlen,value,&l);
    if (rslt >= 0) {
        buffer[outlen] = 0;
        return std::string((char*)buffer);
    } else {
        fprintf(stderr,"Can't convert to latin 1 : %s\n",(const char*)value);
        return std::string((const char*)value);
    }*/
}

extern "C" {

static void my_startElement(	void * ctx, 
                                const xmlChar * name, 
                                const xmlChar ** atts)
{
    XmlParser * xmp = (XmlParser*) ctx;
    XmlParsableItem * item = xmp->current()->newChild(xmlChar2string(name));
    xmp->pushItem(item);
    if (atts) {
	int i = 0;
	while(atts[i]) {
	    item->addAttribute(xmlChar2string(atts[i]),xmlChar2string(atts[i+1]));
	    i+=2;
	}
    }
}

static void my_endElement(	void * ctx,
				const xmlChar * name)
{
    XmlParser * xmp = (XmlParser*) ctx;
    xmp->current()->done();
    if (xmp->current()->getAutoKill()) {
	delete(xmp->current());
    }
    xmp->popItem();
}

static void my_characters(	void *ctx,
				const xmlChar *ch, int len)
{
    XmlParser * xmp = (XmlParser*) ctx;
    xmp->current()->addText(xmlChar2string(ch,len));
}

static xmlEntityPtr my_getEntity(void *user_data, const xmlChar *name) {
    return xmlGetPredefinedEntity(name);
}

}

XmlParser::XmlParser() : stack(),curFileName()
{
    xmlctxt = 0;
    xmlSubstituteEntitiesDefault(1);
}

XmlParser::~XmlParser()
{
    while(!stack.empty()) {
	delete(stack.back());
	stack.pop_back();
    }
}

XmlParsableItem * XmlParser::current() const {
    if (stack.empty()) {
	throw std::logic_error("XmlParser::current called with empty stack");
    }
    return stack.back();
}

void XmlParser::pushItem(XmlParsableItem * newItem)
{
    newItem->setParser(this);
    stack.push_back(newItem);
    newItem->init();
}

void XmlParser::popItem()
{
    stack.pop_back();
}

JSReflection::JSSourceInfo XmlParser::getLocation() const
{
    if (xmlctxt && ((xmlParserCtxtPtr)xmlctxt)->input) {
	return JSReflection::JSSourceInfo(curFileName,((xmlParserCtxtPtr)xmlctxt)->input->line);
    } else {
	return JSReflection::JSSourceInfo(curFileName,0);
    }
}

bool XmlParser::parseString(const std::string & content,const std::string & filename)
{
    xmlSAXHandler xmlParser;
    bool ret;
    
    memset(&xmlParser,0,sizeof(xmlParser));
    xmlParser.getEntity = &my_getEntity;
    xmlParser.startElement = &my_startElement;
    xmlParser.endElement = &my_endElement;
    xmlParser.characters = &my_characters;
   
    xmlParserCtxtPtr ctxt;
    ctxt = xmlCreateMemoryParserCtxt(content.c_str(),content.length());
    if (ctxt == NULL) return false;
   
    curFileName=filename;
    xmlctxt = (void*)ctxt;
    ctxt->sax = &xmlParser;
    ctxt->userData = (void*)this;
   
    xmlParseDocument(ctxt);
    if (ctxt->wellFormed)
        ret = true;
    else
        ret = false;
   
    ctxt->sax = NULL;
   
    xmlFreeParserCtxt(ctxt);
    xmlctxt = 0;
    
    return ret;
}


bool XmlParser::parseDocument(const std::string & filename)
{
    xmlSAXHandler xmlParser;
    bool ret;
    
    memset(&xmlParser,0,sizeof(xmlParser));
    xmlParser.getEntity = &my_getEntity;
    xmlParser.startElement = &my_startElement;
    xmlParser.endElement = &my_endElement;
    xmlParser.characters = &my_characters;

    xmlParserCtxtPtr ctxt;

    ctxt = xmlCreateFileParserCtxt(filename.c_str());
    if (ctxt == NULL) return false;
    
    ctxt->sax = &xmlParser;
    ctxt->userData = (void*)this;

    xmlctxt = (void*)ctxt;
    curFileName = filename;
    
    xmlParseDocument(ctxt);

    if (ctxt->wellFormed)
        ret = true;
    else
        ret = false;
    
    ctxt->sax = NULL;
    
    xmlFreeParserCtxt(ctxt);
    xmlctxt = 0;
    
    return ret;
}

class XmlSimpleRoot: public XmlParsableItem {
protected:
    JSReflection::Object* current;
    int nb_of_root;
    XmlSimpleRoot * reporter;
public:
    std::list<JSReflection::Object*> terminated;
    
    XmlSimpleRoot(int recursiveCount, XmlSimpleRoot * report, JSReflection::Context * c) : XmlParsableItem(c)
    {
        nb_of_root = recursiveCount;
        reporter = (report?report:this);
	current = 0;
	setAutoKill(false);
    }
    
    virtual ~XmlSimpleRoot()
    {
	if (current) { delete(current); }
    }
    
    virtual void addAttribute(const std::string & name,const std::string & value)
    {}
    virtual void addText(const std::string & text)
    {}
    virtual XmlParsableItem * newChild(const std::string & name);
    
    virtual void done()
    {}
    
    void terminate(JSReflection::Object * object)
    {
        reporter->terminated.push_back(object);
        
	current = 0;
    }
};

class XmlSimpleObject : public XmlParsableItem {
protected:
    JSReflection::Object * object;
    XmlSimpleRoot * root;
public:
    XmlSimpleObject(JSReflection::Object * o, XmlSimpleRoot * r, JSReflection::Context * c) : XmlParsableItem(c) {
	object = o;
	root = r;
    }
    
    virtual ~XmlSimpleObject()
    {
    }
    
    virtual void addAttribute(const std::string & name,const std::string & value)
    {
        JSReflection::JS_String str;
        str = value;
        jsval jsv;
        object->GetContext()->ValueToJS( &str, &jsv, "%s", 0 );
	object->setProperty(name.c_str(), jsv);
    }
    
    virtual void addText(const std::string & text)
    {
	// Just ignore.
	//object->setProperty("text",text.c_str());
    }
    
    virtual XmlParsableItem * newChild(const std::string & name)
    {
	JSReflection::Object * child = new JSReflection::Object(getContext());
	object->setProperty(name.c_str(),child);
	return new XmlSimpleObject(child,0,getContext());
    }
    
    virtual void done()
    {
	if (root) {
	    root->terminate(object);
	}
    }
};

XmlParsableItem * XmlSimpleRoot::newChild(const std::string & name)
{
    if (nb_of_root > 0) {
        return new XmlSimpleRoot(nb_of_root - 1, reporter, getContext());    
    } else {

        current = new JSReflection::Object(getContext(), true);
        return new XmlSimpleObject(current,reporter,getContext());
    }
}

typedef struct XmlParserStruct {
    xmlSAXHandler xmlParser;
    xmlParserCtxtPtr ctxt;
} XmlParserStruct;

void XmlParser::start(const std::string & filename,JSReflection::Context * c)
{
    xmlSimpleRoot = new XmlSimpleRoot(1,0,c);
    pushItem((XmlSimpleRoot*)xmlSimpleRoot);
    ownstruct = new XmlParserStruct();

    memset(&((XmlParserStruct*)ownstruct)->xmlParser,0,sizeof(xmlSAXHandler));
    ((XmlParserStruct*)ownstruct)->xmlParser.getEntity = &my_getEntity;
    ((XmlParserStruct*)ownstruct)->xmlParser.startElement = &my_startElement;
    ((XmlParserStruct*)ownstruct)->xmlParser.endElement = &my_endElement;
    ((XmlParserStruct*)ownstruct)->xmlParser.characters = &my_characters;

    ((XmlParserStruct*)ownstruct)->ctxt = xmlCreatePushParserCtxt(&((XmlParserStruct*)ownstruct)->xmlParser,(void*)this,0,0,filename.c_str());
    if (((XmlParserStruct*)ownstruct)->ctxt == NULL) return ;
}

void XmlParser::parse(const char * t, int length,std::list<JSReflection::Object*> & add)
{
     xmlParseChunk(((XmlParserStruct*)ownstruct)->ctxt,t,length,0);

     XmlSimpleRoot * root = (XmlSimpleRoot*)xmlSimpleRoot;
     while(!root->terminated.empty()) {
	 JSReflection::Object * object = *root->terminated.begin();
	 root->terminated.pop_front();
	 add.push_back(object);

     }

}

void  XmlParser::done()
{
    xmlFreeParserCtxt(((XmlParserStruct*)ownstruct)->ctxt);

    delete((XmlSimpleRoot*)xmlSimpleRoot);
}


