/***************************************************************************
*   Copyright (C) 2004 by ludovic p                                       *
*   pludov@nnx.com                                                        *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <stdlib.h>

#include "jsreflect.h"


using namespace JSReflection;

int ClassInfo::Count = 0;
ClassInfo * ClassInfo::First = 0;


ClassInfo::ClassInfo( const char * name,
                      ClassInfo * * ancestor )
{
    Name = name;
    Id = Count++;
    Ancestor = ancestor;
    Next = First;
    First = this;

}

ClassInfo::~ClassInfo()
{}

void ClassInfo::Debug() const
{
    printf("class %s (%s)\n",GetName(),GetAncestor() ? GetAncestor()->GetName() : "n/a");
    FieldDebug();
}

void ClassInfo::FieldDebug() const
{
    const PropertyInfo * prop;
    const FunctionInfo * func;
    if ( GetAncestor() )
    {
        GetAncestor()->FieldDebug();
    }

    prop = GetProperties();
    while ( prop )
    {
        printf( " property %s (%s)\n", prop->Name, ( prop->Set ? "r/w" : "r/o" ) );
        prop = prop->Next;
    }

    func = GetFunctions();
    while ( func )
    {
        func->Debug();
        func = func->Next;
    }
}

const char * ClassInfo::GetName() const
{
    return Name;
}

ClassInfo * ClassInfo::GetAncestor() const
{
    if (Ancestor) {
        return *Ancestor;
    }
    return 0;
}

int ClassInfo::GetId() const
{
    return Id;
}

int ClassInfo::GetClassCount()
{
    return Count;
}

ClassInfo * ClassInfo::FindByName( const char * name )
{
    ClassInfo * current;
    current = First;
    while ( current )
    {
        if ( !strcmp( current->Name, name ) )
        {
            return current;
        }
        current = current->Next;
    }
    
    return 0;
}

bool ClassInfo::SubClassOf(const ClassInfo* m) const
{
    const ClassInfo * cur;
    cur = this;
    while(cur) {
        if (cur==m) {
            return true;
        }
        cur = cur->GetAncestor();
    }
    return false;
}

JSObject * ClassInfo::CreateModel( Context * c ) const
{
    JSObject * model;
    const PropertyInfo * prop;
    const FunctionInfo * func;
    JSPropertySpec propspec[ 2 ];
    JSFunctionSpec funcspec[ 2 ];

    if ( GetAncestor() )
    {
        model = GetAncestor()->CreateModel( c );
    }
    else
    {
        model = c->CreateEmptyModel();
    }

    prop = GetProperties();
    propspec[ 1 ].name = 0;
    propspec[ 1 ].tinyid = 0;
    propspec[ 1 ].flags = 0;
    propspec[ 1 ].getter = 0;
    propspec[ 1 ].setter = 0;

    while ( prop )
    {
        propspec[ 0 ].name = prop->Name;
        propspec[ 0 ].tinyid = 0;
        propspec[ 0 ].flags = JSPROP_ENUMERATE | JSPROP_PERMANENT | ( prop->Set ? 0 : JSPROP_READONLY );
        propspec[ 0 ].getter = prop->Get;
        propspec[ 0 ].setter = prop->Set;

        JS_DefineProperties( c->JSGlobalContext, model, propspec );

        prop = prop->Next;
    }

    func = GetFunctions();
    funcspec[ 1 ].name = 0;
    funcspec[ 1 ].call = 0;
    funcspec[ 1 ].nargs = 0;
    funcspec[ 1 ].flags = 0;
    funcspec[ 1 ].extra = 0;
    while ( func )
    {
        funcspec[ 0 ].name = func->Name;
        funcspec[ 0 ].call = func->Func;
        funcspec[ 0 ].nargs = func->NArgs;
        funcspec[ 0 ].flags = 0;
        funcspec[ 0 ].extra = 0;
        JS_DefineFunctions( c->JSGlobalContext, model, funcspec );
        func = func->Next;
    }

    return model;
}

class ZombiObject:public Object {
public:
    int oldvalue;
    ZombiObject(int id,Context * c): Object(c,false,0) {
	oldvalue = id;
	PreserveCounterpart = false;
    }
    
    virtual bool isDebugZombi() const {
	return true;
    }
    
    void forgetCounterpart() {
	Counterpart = 0;
    }
};

extern "C"
{
    static void JSObject_Finalize(JSContext * cx, JSObject * obj);
};

static void JSObject_Finalize( JSContext * cx, JSObject * obj )
{
    Object * o;

    o = ( Object* ) JS_GetPrivate( cx, obj );

    if ( o )
    {
	if (o->isDebugZombi()) {
	    ZombiObject * zo = (ZombiObject*) o;
	    jsrlog("[%#08x] finalized\n",zo->oldvalue);
	    zo->forgetCounterpart();
	    delete(zo);
	} else {
	    if (o->debugGC()) {
		jsrlog("[%#08x] : finalized but still exists\n",o->debug);
	    }
	    o->SetCounterpart( 0 );
	}
    }
}

void Context::newZombi(JSObject * o,Object * initial) {
    // Warning :doesnot work when deleting
    if (initial->isDebugZombi()) {
	JS_SetPrivate( JSGlobalContext, o, 0 );
	ZombiObject * zo = (ZombiObject*) initial;
	fprintf(stderr,"DEBUGGC : %#x (object %#x) badly finalized\n",o,zo->oldvalue);
	return;
    }
    if (initial->debugGC()) {
	Object * zombi = new ZombiObject(initial->debug,this);
	zombi->SetCounterpart(o);
	JS_SetPrivate( JSGlobalContext, o, (JSObject*)zombi );
    } else {
	JS_SetPrivate( JSGlobalContext, o, 0 );
    }
}

static JSClass EmptyModelClass =
    {
        "Object",
        JSCLASS_HAS_PRIVATE,
        JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
        JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JSObject_Finalize
    };

static JSFunctionSpec EmptyModelFunctions[] = { {0, 0, 0, 0, 0} };

static JSPropertySpec EmptyModelProperties[] = { {0, 0, 0, 0, 0} };

JSObject * Context::CreateEmptyModel()
{
    JSObject * model;
    model = JS_InitClass( JSGlobalContext, JSGlobalObject, JSGlobalObject, &EmptyModelClass,
                          0, 0,
                          EmptyModelProperties,
                          EmptyModelFunctions,
                          0, 0 );
    return model;
}

JSObject * Context::NewInstance( const ClassInfo * cd )
{
    int cdid;
    JSObject * ret;
    cdid = cd->GetId();

    if ( !Models[ cdid ] )
    {
        Models[ cdid ] = cd->CreateModel( this );
    }
    ret = JS_NewObject( JSGlobalContext, &EmptyModelClass, Models[ cdid ], global->Counterpart);
    return ret;
}

JSSourceInfo::JSSourceInfo() : name()
{
    line = 0;
}

JSSourceInfo::JSSourceInfo(const std::string & vname, int lno) :name(vname)
{
    line = lno;
}

JSSourceInfo::~JSSourceInfo()
{
}

std::string JSSourceInfo::getName() const
{
    return name;
}

void JSSourceInfo::setName(const std::string & vname)
{
    name = vname;
}
	
int JSSourceInfo::getLine() const
{
    return line;
}

void JSSourceInfo::setLine(int i)
{
    line = i;
}


JSSourceInfo JSSourceInfo::unknown("unknown",0);


JS_String::JS_String()
{
    NeedFree = 0;
    Value = 0;
}
        
JS_String::~JS_String()
{
    if (NeedFree && Value) {
        
        free((void*)Value);
    }
}

/* When going from parameter -> js */
const JS_String & JS_String::operator=(const char * value)
{
    NeedFree = 0;
    
    Value = value;
    return *this;
}

/* When going from parameter -> js */
const JS_String & JS_String::operator=(const std::string & value)
{
    NeedFree = 1;
    
    Value = strdup(value.c_str());
    return *this;
}

void JS_String::Copy(const char * value, int length)
{
    NeedFree = 1;    
    Value = strndup(value,length);
}

/* When going from js to parameter */
JS_String::operator const char * () const
{
    return Value;
}

/* When going from js to parameter */
JS_String::operator std::string () const
{
    return std::string(Value);
}

const char * JS_String::data() const
{
    return Value;
}

void JS_String::SetFree(bool needfree)
{
    NeedFree = needfree;
}

// convert an utf8 string into a 8bit binary string. Char >255 gets ignored !
std::string JSReflection::decodeBinaryPath(const std::string & utf8)
{
    char buffer[utf8.length() + 1];
    unsigned int i = 0;
    unsigned int length = 0;
    while(i < utf8.length()) {
        unsigned int v;
        unsigned int c = ((unsigned)utf8[i]) & 255;
        i++;
        if (!(c & 128)) {
            v = c;
        } else {
            int length;
            /// 1100 0000, 1110 0000, 11110 0000
            if ((c & 0xE0) == 0xc0) {
                length = 1;
                c &= ~0xE0;
            } else if ((c & 0xF0) == 0xE0) {
                length = 2;
                c &= ~0xF0;
            } else if ((c & 0xF8) == 0xF0) {
                length = 3;
                c &= ~0xF8;
            } else {
                jsrlog("invalid utf8 encoding.\n");
                continue;
            }
            v = c;
            while(length > 0) {

                if (i >= utf8.length()) {
                    jsrlog("invalid utf8 encoding.\n");
                    continue ;
                }
                c = ((unsigned)utf8[i]) & 255;
                ++i;
                length--;
                if ((c & 0xC0) != 0x80) {
                    jsrlog("invalid utf8 encoding.\n");
                    continue ;
                }
                c &= ~ 0xC0;
                v = v << 6 | c;
            }
        }
        buffer[length] = v & 255;
        length++;
    }
    return std::string(buffer,length);
}

std::string JSReflection::displayBinaryPath(const std::string & utf8)
{
    // FIXME : for non utf8 locale, must convert into UTF8 !
    return decodeBinaryPath(utf8);
}


